from rich.console import Console
from rich.terminal_theme import NIGHT_OWLISH as THEME 
from zeep import Plugin
from lxml import etree
import datetime


class RichSoapDebugPlugin(Plugin):
    # timestamp of the last request
    last_request_timestamp = None

    def __init__(self, debug_console: Console = None):
        if debug_console is None:
            self.debug_console = Console(record=True)
        else:
            self.debug_console = debug_console

    def ingress(self, envelope, http_headers, operation):
        elapsed_time = datetime.datetime.now() - self.last_request_timestamp
        self.debug_console.print(f"Response ({elapsed_time}):")
        self.debug_console.print("Headers:")
        for key, value in http_headers.items():
            self.debug_console.print(f"{key}: {value}")
        self.debug_console.print("Body:")
        self.debug_console.print(etree.tostring(envelope, pretty_print=True).decode())

        return envelope, http_headers

    def egress(self, envelope, http_headers, operation, binding_options):
        self.last_request_timestamp = datetime.datetime.now()
        self.debug_console.print(f"SOAP Operation {operation.name}")
        self.debug_console.print("Headers:")
        for key, value in http_headers.items():
            self.debug_console.print(f"{key}: {value}")
        self.debug_console.print("Body:")
        self.debug_console.print(etree.tostring(envelope, pretty_print=True).decode())

        return envelope, http_headers

    def save_html(self, filename: str):
        self.debug_console.save_html(filename, theme=THEME)



from typing import Optional
from urllib.parse import urljoin
from pydantic import BaseSettings, SecretStr
from enum import Enum


class AuthMethod(str, Enum):
    basic = 'basic'
    cert = 'cert'


class ConnectorConfig(BaseSettings):
    base_url: str
    mandant_id: str
    client_system_id: str
    workplace_id: str
    user_id: str
    trustchain: Optional[str] = None

    auth_method: AuthMethod | str = 'basic'
    
    auth_basic_username: Optional[str] = None
    auth_basic_password: Optional[SecretStr] = None

    auth_cert_p12_filename: Optional[str] = None
    auth_cert_p12_password: Optional[SecretStr] = None

    # Danger zone
    danger_verify_tls: bool = True

    def construct_url(self, path: str) -> str:
        return urljoin(self.base_url, path)

    class Config:
        env_prefix = 'konnektor_'



from enum import Enum
from pydantic import BaseModel
from typing import Optional, Mapping, Any
from zeep.helpers import serialize_object


class CardTypeEnum(str, Enum):
    EGK = "EGK"
    HBA_QSIG = "HBA-qSig"
    HBA = "HBA"
    SMC_B = "SMC-B"
    HSM_B = "HSM-B"
    SMC_KT = "SMC-KT"
    KVK = "KVK"
    ZOD_20 = "ZOD_2.0"
    UNKNOWN = "UNKNOWN"
    HBAx = "HBAx"
    SM_B = "SM-B"


CARD_TYPES_SMCB = [
    CardTypeEnum.SMC_B,
    CardTypeEnum.HSM_B,
    CardTypeEnum.SM_B
]


class Card(BaseModel):
    CardType: CardTypeEnum | str
    CardHandle: str
    CardHolderName: str
    Kvnr: Optional[str]
    raw: Optional[Mapping[str, Any]]


def obj_to_card(zeep_obj) -> Card:
    card_dict = serialize_object(zeep_obj)
    card = Card.parse_obj(card_dict)
    card.raw = card_dict
    return card


class CryptEnum(str, Enum):
    RSA = "RSA"
    ECC = "ECC"


class CertRefEnum(str, Enum):
    C_AUT = "C.AUT"
    C_ENC = "C.ENC"
    C_SIG = "C.SIG"
    C_QES = "C.QES"



from lxml.builder import E
from lxml import etree

namespaces = {
    'soap12': 'http://www.w3.org/2003/05/soap-envelope',
    'phrs': 'http://ws.gematik.de/conn/phrs/PHRService/v1.3',
    'phr': 'http://ws.gematik.de/fa/phr/v1.1',
    'conn': 'http://ws.gematik.de/conn/ConnectorContext/v2.0',
    'com': 'http://ws.gematik.de/conn/ConnectorCommon/v5.0',
    'wa': 'http://www.w3.org/2005/08/addressing',
    'xdr': 'urn:ihe:iti:xdr:2014',
    'xdsb': 'urn:ihe:iti:xds-b:2007',
    'lcm': 'urn:oasis:names:tc:ebxml-regrep:xsd:lcm:3.0',
    'rs': 'urn:oasis:names:tc:ebxml-regrep:xsd:rs:3.0',
    'rim': 'urn:oasis:names:tc:ebxml-regrep:xsd:rim:3.0',
    'query': 'urn:oasis:names:tc:ebxml-regrep:xsd:query:3.0',
}

# define namespace prefixes at library level for better readability
for prefix, uri in namespaces.items():
    etree.register_namespace(prefix, uri)


def Soap12Envelope(*children, **attrib):
    return E('{http://www.w3.org/2003/05/soap-envelope}Envelope', *children, **attrib)


def Soap12Header(*children, **attrib):
    return E('{http://www.w3.org/2003/05/soap-envelope}Header', *children, **attrib)


def Soap12Body(*children, **attrib):
    return E('{http://www.w3.org/2003/05/soap-envelope}Body', *children, **attrib)


def ContextHeader(*children, **attrib):
    return E('{http://ws.gematik.de/conn/phrs/PHRService/v1.3}ContextHeader', *children, **attrib)


def Context(context: dict):
    return E(
        '{http://ws.gematik.de/conn/ConnectorContext/v2.0}Context',
        E('{http://ws.gematik.de/conn/ConnectorCommon/v5.0}MandantId', context['MandantId']),
        E('{http://ws.gematik.de/conn/ConnectorCommon/v5.0}ClientSystemId', context['ClientSystemId']),
        E('{http://ws.gematik.de/conn/ConnectorCommon/v5.0}WorkplaceId', context['WorkplaceId']),
    )


def RecordIdentifier(*children, **attrib):
    return E('{http://ws.gematik.de/conn/phrs/PHRService/v1.3}RecordIdentifier', *children, **attrib)


def InsurantId(*children, **attrib):
    return E('{http://ws.gematik.de/fa/phr/v1.1}InsurantId', *children, **attrib)


def HomeCommunityId(*children, **attrib):
    return E('{http://ws.gematik.de/fa/phr/v1.1}HomeCommunityId', *children, **attrib)


def Action(*children, **attrib):
    return E('{http://www.w3.org/2005/08/addressing}Action', *children, **attrib)


# <MessageID xmlns="http://www.w3.org/2005/08/addressing">6e2fa9e2-18fb-4071-b796-a49c5fe9a303</MessageID>
def MessageID(*children, **attrib):
    return E('{http://www.w3.org/2005/08/addressing}MessageID', *children, **attrib)


# <ReplyTo xmlns="http://www.w3.org/2005/08/addressing">
def ReplyTo(*children, **attrib):
    return E('{http://www.w3.org/2005/08/addressing}ReplyTo', *children, **attrib)


# <Address>http://www.w3.org/2005/08/addressing/anonymous</Address>
def Address(*children, **attrib):
    return E('{http://www.w3.org/2005/08/addressing}Address', *children, **attrib)


# <To xmlns="http://www.w3.org/2005/08/addressing">https://192.168.1.194:443/ws/PHRManagementService</To>
def To(*children, **attrib):
    return E('{http://www.w3.org/2005/08/addressing}To', *children, **attrib)


# <xdr:homeCommunityBlock xmlns:xdr="urn:ihe:iti:xdr:2014">
def homeCommunityBlock(*children, **attrib):
    return E('{urn:ihe:iti:xdr:2014}homeCommunityBlock', *children, **attrib)


# <xdr:homeCommunityId>urn:oid:1.2.276.0.76.3.1.315.3.3.1.1</xdr:homeCommunityId>
def homeCommunityId(*children, **attrib):
    return E('{urn:ihe:iti:xdr:2014}homeCommunityId', *children, **attrib)


def ProvideAndRegisterDocumentSetRequest(*children, **attrib):
    return E('{urn:ihe:iti:xds-b:2007}ProvideAndRegisterDocumentSetRequest', *children, **attrib)


def SubmitObjectsRequest(*children, **attrib):
    return E('{urn:oasis:names:tc:ebxml-regrep:xsd:lcm:3.0}SubmitObjectsRequest', *children, **attrib)


# <rs:RequestSlotList>
def RequestSlotList(*children, **attrib):
    return E('{urn:oasis:names:tc:ebxml-regrep:xsd:rs:3.0}RequestSlotList', *children, **attrib)


# <rim:Slot name="homeCommunityId">
def Slot(*children, **attrib):
    return E('{urn:oasis:names:tc:ebxml-regrep:xsd:rim:3.0}Slot', *children, **attrib)


# <rim:ValueList>
def ValueList(*children, **attrib):
    return E('{urn:oasis:names:tc:ebxml-regrep:xsd:rim:3.0}ValueList', *children, **attrib)


# <rim:Value>urn:oid:1.2.276.0.76.3.1.315.3.3.1.1</rim:Value>
def Value(*children, **attrib):
    return E('{urn:oasis:names:tc:ebxml-regrep:xsd:rim:3.0}Value', *children, **attrib)


# <rim:RegistryObjectList>
def RegistryObjectList(*children, **attrib):
    return E('{urn:oasis:names:tc:ebxml-regrep:xsd:rim:3.0}RegistryObjectList', *children, **attrib)


# <rim:RegistryPackage id="submissionset" objectType="urn:oasis:names:tc:ebxml-regrep:ObjectType:RegistryObject:RegistryPackage" home="urn:oid:1.2.276.0.76.3.1.315.3.3.1.1">
def RegistryPackage(*children, **attrib):
    return E('{urn:oasis:names:tc:ebxml-regrep:xsd:rim:3.0}RegistryPackage', *children, **attrib)


# <rim:Classification classificationScheme="urn:uuid:a7058bb9-b4e4-4307-ba5b-e3f0ab85e12d" classifiedObject="submissionset" id="author" objectType="urn:oasis:names:tc:ebxml-regrep:ObjectType:RegistryObject:Classification">
def Classification(*children, **attrib):
    return E('{urn:oasis:names:tc:ebxml-regrep:xsd:rim:3.0}Classification', *children, **attrib)


# <rim:Name>
def Name(*children, **attrib):
    return E('{urn:oasis:names:tc:ebxml-regrep:xsd:rim:3.0}Name', *children, **attrib)


# <rim:LocalizedString xml:lang="de-DE" value="Veranlassung durch Patient"/>
def LocalizedString(*children, **attrib):
    return E('{urn:oasis:names:tc:ebxml-regrep:xsd:rim:3.0}LocalizedString', *children, **attrib)


# <rim:ExternalIdentifier id="patientId" identificationScheme="urn:uuid:6b5aea1a-874d-4603-a4bc-96a0a7b38446" objectType="urn:oasis:names:tc:ebxml-regrep:ObjectType:RegistryObject:ExternalIdentifier" registryObject="submissionset" value="X110476138^^^&1.2.276.0.76.4.8&ISO">
def ExternalIdentifier(*children, **attrib):
    return E('{urn:oasis:names:tc:ebxml-regrep:xsd:rim:3.0}ExternalIdentifier', *children, **attrib)


# <rim:Association associationType="urn:oasis:names:tc:ebxml-regrep:AssociationType:HasMember" id="association-0" sourceObject="submissionset" targetObject="DocumentEntry-0">
def Association(*children, **attrib):
    return E('{urn:oasis:names:tc:ebxml-regrep:xsd:rim:3.0}Association', *children, **attrib)


# <rim:ExtrinsicObject id="DocumentEntry-0" mimeType="application/xml" objectType="urn:uuid:7edca82f-054d-47f2-a032-9b2a5b5186c1" home="urn:oid:1.2.276.0.76.3.1.315.3.3.1.1">
def ExtrinsicObject(*children, **attrib):
    return E('{urn:oasis:names:tc:ebxml-regrep:xsd:rim:3.0}ExtrinsicObject', *children, **attrib)


# <Document id="DocumentEntry-0">
def Document(*children, **attrib):
    return E('{urn:ihe:iti:xds-b:2007}Document', *children, **attrib)


# <Include xmlns="http://www.w3.org/2004/08/xop/include" href="cid:Document0@PHRService.konlan"/></Document>
def Include(*children, **attrib):
    return E('{http://www.w3.org/2004/08/xop/include}Include', *children, **attrib)


# <query:AdhocQueryRequest xmlns:ns0="urn:oasis:names:tc:ebxml-regrep:xsd:query:3.0" federated="false" startIndex="0" maxResults="-1">
def AdhocQueryRequest(*children, **attrib):
    return E('{urn:oasis:names:tc:ebxml-regrep:xsd:query:3.0}AdhocQueryRequest', *children, **attrib)


# <query:ResponseOption returnType="LeafClass" returnComposedObjects="true"/>
def ResponseOption(*children, **attrib):
    return E('{urn:oasis:names:tc:ebxml-regrep:xsd:query:3.0}ResponseOption', *children, **attrib)


# <ns1:AdhocQuery xmlns:ns1="urn:oasis:names:tc:ebxml-regrep:xsd:rim:3.0" id="urn:uuid:958f3006-baad-4929-a4de-ff1114824431" home="urn:oid:1.2.276.0.76.3.1.315.3.3.1.1">
def AdhocQuery(*children, **attrib):
    return E('{urn:oasis:names:tc:ebxml-regrep:xsd:rim:3.0}AdhocQuery', *children, **attrib)



from zeep.xsd.elements.element import Element
from zeep.xsd.elements import Attribute
from typing import Mapping, List, Set
from zeep.xsd.types import ComplexType, AnySimpleType
from zeep.xsd.elements import Sequence, Choice
from zeep.xsd.elements.any import Any
from zeep.wsdl.bindings.soap import Soap11Binding
from zeep import Client, Settings
from datetime import datetime
import os
from urllib.parse import urlparse
import logging
import re
from pathlib import Path
from importlib import resources
from dataclasses import dataclass


shortnames = {
  "urn:oasis:names:tc:dss:1.0:core:schema": "ext/dss10core",
  "urn:oasis:names:tc:dss-x:1.0:profiles:verificationreport:schema#": "ext/dssx10verificationreport",
  "http://www.w3.org/2000/09/xmldsig#": "ext/xmldsig",
  "http://uri.etsi.org/02231/v2#": "ext/etsi02232v2",
  "http://uri.etsi.org/01903/v1.3.2#": "ext/etsi01903v132",
}


@dataclass
class ServiceBindingDefinition:
    name: str
    version: str
    soap11Binding: Soap11Binding

    def types_namespace(self):
        return self.soap11Binding.port_type.name.namespace.replace("/WSDL/v", "/v")


@dataclass
class ElementDefinition:
    element: Element
    version: str = None

    def __hash__(self):
        return hash((self.element.qname.namespace, self.element.qname.localname, self.version))
    
    def __eq__(self, other):
        return (
            self.element.qname.namespace == other.element.qname.namespace
            and self.element.qname.localname == other.element.qname.localname
            and self.version == other.version
        )


class GoGenerator:
    def __init__(self, dest: str, package: str):
        self.dest = dest
        self.package = package
        os.makedirs(dest, exist_ok=True)

    elements: Set[ElementDefinition] = set()
    bindings: List[ServiceBindingDefinition] = []

    def scan_for_wsdl(self):
        wsdl_files = resources.files("koap") / "api-telematik" / "conn"
        # we make use of gematik naming conventions:
        # - the wsdl file name is the service name
        # - if wsdl file name has vX_Y_Z.wsdl, then the service version is vX.Y.Z
        # - if no version is pecified in the wsdl file name, then the service version taken from targetNamespace. Patch level is 0
        for wsdl_file in wsdl_files.glob("*.wsdl"):
            service_name = wsdl_file.stem
            match = re.match(r"^([^\_]+)_v(\d+_\d+_\d+)$", service_name, re.IGNORECASE)
            if match:
                service_name = match.group(1)
                service_version = match.group(2).replace("_", ".")
            else:
                soap_settings = Settings()
                soap_settings.forbid_entities = False
                client = Client(wsdl=str(wsdl_file), settings=soap_settings)
                _, binding = list(client.wsdl.bindings.items())[0]
                namespace = binding.name.namespace
                service_version = namespace.split("/")[-1].replace("v", "") + ".0"

            self.add_service(wsdl_file, service_name, service_version)

    def add_service(
        self,
        wsdl_file: str,
        service_name: str,
        service_version: str,
    ):

        soap_settings = Settings()
        soap_settings.forbid_entities = False

        client = Client(
            wsdl=str(wsdl_file),
            settings=soap_settings,
        )

        for svc in client.wsdl.services.values():
            for port in svc.ports.values():
                binding = port.binding
                if isinstance(binding, Soap11Binding):
                    service_binding = ServiceBindingDefinition(
                        name=service_name,
                        version=service_version,
                        soap11Binding=binding,
                    )
                    self.bindings.append(service_binding)

                    for op in port.binding.all().values():
                        self.add_type(service_binding.types_namespace(), op.abstract.input_message.parts["parameter"].element, version=service_version)
                        self.add_type(service_binding.types_namespace(), op.abstract.output_message.parts["parameter"].element, version=service_version)

    def add_type(self, context_namespase: str, element: Element, version: str = None):
        ns = element.qname.namespace

        if ns != context_namespase:
            version = None

        element_definition = ElementDefinition(element=element, version=version)

        if element_definition in self.elements:
            # already added
            return

        self.elements.add(element_definition)

        if isinstance(element.type, ComplexType):
            for child in self._iterate_children(element):
                if isinstance(child, Element) and isinstance(child.type, ComplexType):
                    self.add_type(context_namespase, child, version)
                elif isinstance(child, Element) and isinstance(child.type, AnySimpleType):
                    # TODO: handle simple types
                    pass
                else:
                    logging.warning(
                        f"Skipping unknown {type(child.type)} {type(child)}"
                    )
        else:
            logging.warning(
                f"Not complex type {type(element.type)} {element.type}"
            )

    # given a namespace, return a list of paths to the types in that namespace
    def paths(self, ns: str, version: str = None) -> List[str]:
        if ns in shortnames:
            ns = shortnames[ns]
        # parse url
        url = urlparse(ns)
        # split path
        paths = url.path.split("/")
        # remove empty elements
        paths = [p for p in paths if p]
        # if last path is semantic version and version in None, prepend with _v
        if re.match(r"v\d+\.\d+(\.\d+)?", paths[-1]):
            if version is None:
                paths[-2] = paths[-2] + "_" + paths[-1].replace(".", "_")
            else:
                paths[-2] = paths[-2] + "_v" + version.replace(".", "_")

            paths.pop()

        return paths

    def generate_all(self):
        self.generate_all_types()
        self.generate_all_bindings()

    def generate_all_types(self):
        # group types by namespace and version
        elements_by_file: Mapping[str, List[Element]] = {}
        for element in self.elements:
            dest_dir = Path(self.dest, *self.paths(element.element.qname.namespace, element.version))
            os.makedirs(dest_dir, exist_ok=True)
            dest_file = dest_dir / "types.gen.go"
            if dest_file not in elements_by_file:
                elements_by_file[dest_file] = []
            elements_by_file[dest_file].append(element.element)

        for dest_file, elements in elements_by_file.items():
            self.generate_complextypes(dest_file, elements)

    def generate_complextypes(self, dest_file: Path, elements: List[Element]):
        logging.info(f"Generating {len(elements)} complex types to {dest_file}")
        elements = sorted(elements, key=lambda e: e.qname.localname)

        file_imports = set()
        file_imports.add("encoding/xml")

        file_code = ""
        for element in elements:
            if isinstance(element.type, ComplexType):
                code, imports = self._element_to_struct(element)
                file_code += code
                file_code += "\n"
                file_imports.update(imports)

        if file_code == "":
            # do not generate empty files
            return
        with open(dest_file, "w") as f:
            f.write("package " + dest_file.parent.name + "\n\n")
            f.write("import (\n")
            for imp in file_imports:
                f.write(f'    "{imp}"\n')
            f.write(")\n\n")
            f.write(file_code)

    def generate_all_bindings(self):
        for binding in self.bindings:
            self.generate_binding(binding)

    def generate_binding(self, binding: ServiceBindingDefinition):
        ns = binding.types_namespace()
        paths = self.paths(ns, binding.version)

        dest_dir = Path(self.dest, *paths)
        os.makedirs(dest_dir, exist_ok=True)
        dest_file = dest_dir / "binding.gen.go"

        logging.info(f"Generating binding to {dest_file}")

        file_imports = set()
        file_imports.add("reflect")
        file_imports.add("github.com/spilikin/koap-go/pkg/koap")
        file_code = ""

        for op_name in binding.soap11Binding.port_type.operations:
            op = binding.soap11Binding.get(op_name)
            code, imports = self._operation_to_go(binding.types_namespace(), op)
            file_code += code
            file_imports.update(imports)

        with open(dest_file, "w") as f:
            f.write("package " + dest_dir.name + "\n\n")
            f.write("import (\n")
            for imp in file_imports:
                f.write(f'    "{imp}"\n')
            f.write(")\n\n")
            f.write(file_code)

    def _operation_to_go(self, types_namespace, op) -> tuple[str, Set[str]]:
        imports = set()

        inputType, inputImport = self._go_type_of(types_namespace, op.input.body)
        if inputImport is not None:
            imports.add(inputImport)
        outputType, outputImport = self._go_type_of(types_namespace, op.output.body)
        if outputImport is not None:
            imports.add(outputImport)

        faultElement = op.faults['FaultMessage'].abstract.parts['parameter'].element
        faultType, pkg = self._go_type_of(op.input.body.qname.namespace, faultElement)

        if pkg is not None:
            imports.add(pkg)

        s = f'var Operation{op.name} = &koap.Operation{{\n'
        s += f'    Name: "{op.name}",\n'
        s += f'    SOAPAction: "{op.soapaction}",\n'
        s += f'    InputType: reflect.TypeOf({inputType}{{}}),\n'
        s += f'    OutputType: reflect.TypeOf({outputType}{{}}),\n'
        s += f'    FaultType: reflect.TypeOf({faultType}{{}}),\n'
        s += "}\n\n"
        return s, imports

    def _iterate_children(self, element: ComplexType):
        for name, child in element.type.elements_nested:
            if isinstance(child, Sequence) or isinstance(child, Choice):
                # TODO: handle sequence and choice
                processed = set()
                for name, grandchild in child.elements:
                    if isinstance(grandchild, Any):
                        continue
                    if grandchild.qname in processed:
                        continue
                    processed.add(grandchild.qname)
                    yield grandchild
            elif isinstance(child, Element):
                yield child
            else:
                logging.warning(f"Unknown type: {type(child)} in {element.type}")

    def _element_to_struct(self, element: ComplexType) -> tuple[str, Set[str]]:
        imports = set()
        s = f"type {element.name} struct {{\n"
        s += f'XMLName        xml.Name `xml:"{element.qname.namespace} {element.qname.localname}"`\n'

        for _, attr in element.type.attributes:
            opts = ",attr"

            if not isinstance(attr, Attribute):
                logging.warning(f"Skipping {type(attr)}: {attr}")
                continue

            go_field = self._go_field(attr.qname.localname)
            go_type, imp = self._go_type_of(element.qname.namespace, attr)
            if imp is not None:
                imports.add(imp)
            s += f'    {go_field} {go_type} `xml:"{attr.qname.localname}{opts}"`\n'

        for child in self._iterate_children(element):
            opts = ""
            if isinstance(child, Any):
                continue

            element_name = f"{child.qname.namespace} {child.qname.localname}"
            go_field = self._go_field(child.qname.localname)
            if child.qname.localname == "_value_1":
                go_field = "Value"
                element_name = ""
                opts = ",chardata"

            go_type, imp = self._go_type_of(element.qname.namespace, child)
            # see if go type is from another package (contains .)
            if imp is not None:
                imports.add(imp)
            if child.min_occurs == 0:
                opts += ",omitempty"
            if child.max_occurs == "unbounded" or int(child.max_occurs) > 1:
                go_type = "[]" + go_type
            s += f'    {go_field} {go_type} `xml:"{element_name}{opts}"`\n'
        s += "}\n"

        return s, imports

    def _go_type_of(self, current_namespace: str, child: Element) -> tuple[str, str]:
        if isinstance(child.type, ComplexType):
            if current_namespace == child.qname.namespace:
                return child.qname.localname, None
            else:
                paths = self.paths(child.qname.namespace)
                go_type = paths[-1] + "." + child.qname.localname
                go_package = self.package + "/" + "/".join(paths)
                return go_type, go_package
        elif len(child.type.accepted_types) > 0:
            pyt = child.type.accepted_types[0]
            if pyt == str:
                return "string", None
            elif pyt == int:
                return "int", None
            elif pyt == float:
                return "float64", None
            elif pyt == bool:
                return "bool", None
            elif pyt == datetime:
                return "time.Time", "time"
            else:
                return "interface{}", None
        else:
            return "interface{}", None

    def _go_field(self, localname: str):
        # convert snake case to camel case
        parts = localname.split("_")
        localname = parts[0] + "".join(x.title() for x in parts[1:])
        parts = localname.split("-")
        localname = parts[0] + "".join(x.title() for x in parts[1:])
        remove = re.compile(r"[^a-zA-Z0-9_]")
        localname = remove.sub("", localname)
        return localname[0].upper() + localname[1:]



from typing import Any, List, Mapping
from .epa_facade import EPAFacade, RecordId
from lxml import etree
import xmltodict
from uuid import uuid4
from . import builder
from zeep.exceptions import TransportError, Fault


# very quick and dirty hack to remove the namespace prefix and convert to dict
def etree_to_dict(el: etree.ElementBase):
    for el2 in el.findall('.//*'):
        if el2.tag.startswith('{'):
            el2.tag = el2.tag.split('}', 1)[1]

    as_dist = xmltodict.parse(etree.tostring(el, encoding='utf8', method='xml', pretty_print=True))

    # get the value of the fist entry in dict
    as_dist = as_dist[list(as_dist.keys())[0]]

    return as_dist


class EPARecordFacade:
    def __init__(self, epa_facade: EPAFacade, record_id: RecordId):
        self.epa_facade = epa_facade
        self.record_id = record_id
        self.session = epa_facade.phr_service._client.transport.session
        self.address: str = epa_facade.phr_service._binding_options['address']
        self.plugins = epa_facade.phr_service._client.plugins
        if self.plugins is None:
            self.plugins = []

    def find_folders(self) -> List[Mapping[str, Any]]:
        soap_body = builder.Soap12Body(
            builder.AdhocQueryRequest(
                builder.ResponseOption(returnType="LeafClass", returnComposedObjects="true"),
                builder.AdhocQuery(
                    builder.Slot(
                        builder.ValueList(
                            builder.Value("('urn:oasis:names:tc:ebxml-regrep:StatusType:Approved')")
                        ),
                        name="$XDSFolderStatus"
                    ),
                    builder.Slot(
                        builder.ValueList(
                            builder.Value(f"('{self.record_id.InsurantId.extension}^^^&amp;1.2.276.0.76.4.8&amp;ISO')")
                        ),
                        name="$XDSFolderPatientId"
                    ),
                    id="urn:uuid:958f3006-baad-4929-a4de-ff1114824431",
                    home=self.record_id.HomeCommunityId,
                ),
                federated="false",
                startIndex="0",
                maxResults="-1",
            )
        )

        query_response = self.soap12_call('DocumentRegistry_RegistryStoredQuery', soap_body)

        return query_response['AdhocQueryResponse']['RegistryObjectList']['RegistryPackage']

    def get_folders_and_contents(self, folder_uuid) -> Mapping[str, Any]:
        soap_body = builder.Soap12Body(
            builder.AdhocQueryRequest(
                builder.ResponseOption(returnType="LeafClass", returnComposedObjects="true"),
                builder.AdhocQuery(
                    builder.Slot(
                        builder.ValueList(
                            builder.Value(f"'{folder_uuid}'")
                        ),
                        name="$XDSFolderEntryUUID"
                    ),
                    id="urn:uuid:b909a503-523d-4517-8acf-8e5834dfc4c7",
                    home=self.record_id.HomeCommunityId,
                ),
                federated="false",
                startIndex="0",
                maxResults="-1",
            )
        )

        query_response = self.soap12_call('DocumentRegistry_RegistryStoredQuery', soap_body)

        return query_response['AdhocQueryResponse']['RegistryObjectList']

    def find_documents(self) -> Any:
        soap_body = builder.Soap12Body(
            builder.AdhocQueryRequest(
                builder.ResponseOption(returnType="LeafClass", returnComposedObjects="true"),
                builder.AdhocQuery(
                    builder.Slot(
                        builder.ValueList(
                            builder.Value("('urn:oasis:names:tc:ebxml-regrep:StatusType:Approved')")
                        ),
                        name="$XDSDocumentEntryStatus"
                    ),
                    builder.Slot(
                        builder.ValueList(
                            builder.Value(f"('{self.record_id.InsurantId.extension}^^^&amp;1.2.276.0.76.4.8&amp;ISO')")
                        ),
                        name="$XDSDocumentEntryPatientId"
                    ),
                    id="urn:uuid:14d4debf-8f97-4251-9a74-a90016b0af0d",
                    home=self.record_id.HomeCommunityId,
                ),
                federated="false",
                startIndex="0",
                maxResults="-1",
            )
        )

        query_response = self.soap12_call('DocumentRegistry_RegistryStoredQuery', soap_body)

        return query_response['AdhocQueryResponse']['RegistryObjectList']['ExtrinsicObject']

    def soap_http_headers(self, soap_action: str) -> Mapping[str, str]:
        return {
            'Content-Type': f'application/soap+xml; charset=utf-8; action="{soap_action}"',
            # 'SOAPAction': soap_action,
        }

    def soap_header(self, soap_action: str) -> etree.Element:
        messageID = uuid4()
        return builder.Soap12Header(
            builder.ContextHeader(
                builder.Context(self.epa_facade.client.context()),
                builder.RecordIdentifier(
                    builder.InsurantId(root='1.2.276.0.76.4.8', extension=self.record_id.InsurantId.extension),
                    builder.HomeCommunityId(self.record_id.HomeCommunityId)
                ),
            ),
            builder.Action(soap_action),
            builder.MessageID(str(messageID)),
            builder.To(self.address),
            builder.ReplyTo(
                builder.Address('http://www.w3.org/2005/08/addressing/anonymous')
            ),
            builder.homeCommunityBlock(
                builder.homeCommunityId(self.record_id.HomeCommunityId)
            )
        )

    def soap12_call(self, operation_name: str, soap_body: etree.Element) -> etree.Element:
        # Zeep SOAP Operation we will be calling manuelly
        operation = self.epa_facade.phr_service._binding._operations[operation_name]
        # SOAP Action from Zeep Operation
        soap_action = operation.soapaction

        # prepare SOAP Envelpe XML Element
        message = builder.Soap12Envelope(
            self.soap_header(soap_action),
            soap_body
        )

        # Prepare HTTP Headers
        http_headers = self.soap_http_headers(soap_action)

        # Notify plugins, that request is incoming
        for plugin in self.plugins:
            plugin.egress(message, http_headers, operation, None)

        # serialize SOAP Envelope to XML string
        post_data = etree.tostring(message, pretty_print=True).decode()

        # perform HTTP POST to WebService
        response = self.session.post(
            self.address,
            headers=http_headers,
            data=post_data
        )

        if response.status_code != 200:
            raise TransportError(f"Server returned unexpected status: {response.status_code}", status_code=response.status_code, content=response.content)

        response_envelope = etree.fromstring(response.content)

        # Notify plugins, that response is received
        for plugin in self.plugins:
            plugin.ingress(response_envelope, http_headers, operation)

        # Check for SOAP Fault
        fault = response_envelope.xpath('/soap12:Envelope/soap12:Body/soap12:Fault', namespaces=builder.namespaces)
        if len(fault) > 0:
            raise Fault(fault[0].xpath('soap12:Reason/soap12:Text/text()', namespaces=builder.namespaces)[0])

        # Check for RegistryResponse Errors
        query_response_status = response_envelope.xpath('/soap12:Envelope/soap12:Body/query:AdhocQueryResponse/@status', namespaces=builder.namespaces)
        if len(query_response_status) > 0 and query_response_status[0] == 'urn:oasis:names:tc:ebxml-regrep:ResponseStatusType:Failure':
            errors = response_envelope.xpath('/soap12:Envelope/soap12:Body/query:AdhocQueryResponse/rs:RegistryErrorList/rs:RegistryError/@codeContext', namespaces=builder.namespaces)
            raise Fault(','.join(errors))

        return etree_to_dict(response_envelope.xpath('/soap12:Envelope/soap12:Body', namespaces=builder.namespaces)[0])


def slot_by_name(obj: dict, name: str):
    for slot in obj.get('Slot', []):
        if slot['@name'] == name:
            return slot
    raise KeyError(f"Slot with name '{name}' not found")


def classification_slot_by_name(obj: dict, name: str):
    for classification in obj.get('Classification', []):
        for slot in classification.get('Slot', []):
            if slot['@name'] == name:
                return slot
    raise KeyError(f"Slot with name '{name}' not found")



from koap.client import ConnectorClient
from pydantic import BaseModel
from koap.facade.model import CardTypeEnum
from koap.facade.model import obj_to_card, Card, CARD_TYPES_SMCB
from typing import Tuple, List
import datetime


class InsurantId(BaseModel):
    extension: str
    root: str = "1.2.276.0.76.4.8"

    def __init__(self, kvnr: str):
        super().__init__(extension=kvnr)


class RecordId(BaseModel):
    HomeCommunityId: str
    InsurantId: InsurantId


class EPAFacade:
    def __init__(self, client: ConnectorClient):
        self.client = client
        self.phr_mgt_service = client.create_service_client('PHRManagementService', '2.0.1', module="phrs")
        self.phr_service = client.create_service_client('PHRService', '2.0.1', module="phrs", binding_local_name='PHRService_Binding_Soap12')
        self.event_service = client.create_service_client('EventService', '7.2.0')

    def get_home_community_id(self, kvnr: str):
        response = self.phr_mgt_service.GetHomeCommunityID(
            Context=self.client.context(),
            InsurantID=InsurantId(extension=kvnr).dict(),
        )
        # TODO: camelcase ID
        return response['HomeCommunityID']

    def get_authorization_list(self):
        response = self.phr_mgt_service.GetAuthorizationList(
            Context=self.client.context(),
        )
        return response['AuthorizationList']['_value_1']

    def get_cards(self) -> Tuple[List[Card], List[Card]]:
        get_cards_response = self.event_service.GetCards(
            Context=self.client.context()
        )
        raw_egks = filter(lambda c: c.CardType in [CardTypeEnum.EGK], get_cards_response.Cards.Card)
        raw_smcbs = filter(lambda c: c.CardType in CARD_TYPES_SMCB, get_cards_response.Cards.Card)
        return list(map(obj_to_card, raw_egks)), list(map(obj_to_card, raw_smcbs))

    def request_facility_authorization(
            self,
            egk: Card,
            smcb: Card,
            record_id: RecordId,
            categories: List[str],
            expiration_date: datetime.date
            ):
        response = self.phr_mgt_service.RequestFacilityAuthorization(
            Context=self.client.context(),
            EhcHandle=egk.CardHandle,
            # TODO: woher weiss man das?
            AuthorizationConfiguration={
                'AuthorizationConfidentiality': 'normal',
                'DocumentCategoryList': {
                    '_value_1': list(map(lambda c: {'DocumentCategoryElement': c}, categories)),
                },
                # TODO: set date
                'ExpirationDate': expiration_date
            },
            RecordIdentifier=record_id.dict(),
            OrganizationName=smcb.CardHolderName,
            # TODO: Vorname und Nachname: Deutsch(!) wo kommt das her?
            InsurantName={"Vorname": egk.CardHolderName, "Nachname": egk.CardHolderName},
        )
        return response

    def get_authorization_state(self, record_id: RecordId):
        response = self.phr_mgt_service.GetAuthorizationState(
            Context=self.client.context(),
            RecordIdentifier=record_id.dict(),
            # TODO: get koap name and version
            UserAgent="koap/0.1.0/spilikin",
        )
        return response



from koap.client import ConnectorClient
from koap.facade.model import (
    CardTypeEnum,
    Card,
    obj_to_card,
    CryptEnum,
    CertRefEnum,
)
from typing import List, Optional
from pydantic import BaseModel
import gzip
import xmltodict
from zeep.helpers import serialize_object


class VSD(BaseModel):
    PersoenlicheVersichertendaten: dict
    AllgemeineVersicherungsdaten: dict
    GeschuetzteVersichertendaten: dict
    Pruefungsnachweis: Optional[dict]
    VSD_Status: Optional[dict]


class VSDMFacade:
    def __init__(self, client: ConnectorClient):
        self.client = client
        self.event_service = client.create_service_client('EventService', '7.2.0')
        self.vsd_service = client.create_service_client('VSDService', '5.2.0', module="vsds")        

    def get_cards(self, types: List[CardTypeEnum] = None) -> List[Card]:
        get_cards_response = self.event_service.GetCards(
            Context=self.client.context()
        )
        raw_cards = get_cards_response.Cards.Card
        if types is not None:
            raw_cards = filter(lambda c: c.CardType in types, raw_cards)
        return list(map(obj_to_card, raw_cards))

    def get_card_certificates(self, card_handle: str, cert_types: List[CertRefEnum], crypt: CryptEnum):
        response = self.certificate_service.ReadCardCertificate(
            CardHandle=card_handle,
            Context=self.client.context(),
            CertRefList=list(map(lambda cert_type: cert_type.value, cert_types)),
            Crypt=crypt.value,
        )

        return response.X509DataInfoList.X509DataInfo

    def read_vsd(self, EhcHandle: str, HpcHandle: str, PerformOnlineCheck: bool, ReadOnlineReceipt: bool) -> VSD:
        response = self.vsd_service.ReadVSD(
            Context=self.client.context(),
            EhcHandle=EhcHandle,
            HpcHandle=HpcHandle,
            PerformOnlineCheck=PerformOnlineCheck,
            ReadOnlineReceipt=ReadOnlineReceipt,
        )

        return VSD(
            PersoenlicheVersichertendaten=self.vsd_data_to_dict(response, 'PersoenlicheVersichertendaten'),
            AllgemeineVersicherungsdaten=self.vsd_data_to_dict(response, 'AllgemeineVersicherungsdaten'),
            GeschuetzteVersichertendaten=self.vsd_data_to_dict(response, 'GeschuetzteVersichertendaten'),
            Pruefungsnachweis=self.vsd_data_to_dict(response, 'Pruefungsnachweis'),
            VSD_Status=serialize_object(response['VSD_Status']),
        )

    def vsd_data_to_dict(self, response: dict, field_name: str) -> dict:
        compressed_data = response[field_name]
        if compressed_data is None:
            return None
        # gunzip comnpresed data
        data = gzip.decompress(compressed_data)
        xml_as_dict = xmltodict.parse(data)
        # first value in the dict
        xml_as_dict = xml_as_dict[list(xml_as_dict.keys())[0]]
        return xml_as_dict


import requests
import base64
import xml.etree.ElementTree as ET
from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization


TSL_URL_TU = "https://download-test.tsl.ti-dienste.de/ECC/ECC-RSA_TSL-test.xml"
TSL_URL_RU = "https://download-ref.tsl.ti-dienste.de/ECC/ECC-RSA_TSL-ref.xml"
TSL_URL_PU = "https://download.tsl.ti-dienste.de/ECC/ECC-RSA_TSL.xml"

TYPE_CA_NON_QES = "http://uri.etsi.org/TrstSvc/Svctype/CA/PKC"


def tsl_to_pem(tsl_url: str, output_file: str):
    response = requests.get(tsl_url)

    if response.status_code != 200:
        raise Exception("Failed to retrieve XML. Error code:", response.status_code)

    root = ET.fromstring(response.content)

    ns = {'tsl': 'http://uri.etsi.org/02231/v2#'}

    with open(output_file, "wb") as out:
        # Iterate over the elements
        for element in root.findall('./tsl:TrustServiceProviderList/tsl:TrustServiceProvider/tsl:TSPServices/tsl:TSPService', ns):
            type = element.findtext('./tsl:ServiceInformation/tsl:ServiceTypeIdentifier', namespaces=ns)
            if type != TYPE_CA_NON_QES:
                continue
            # TODO: differentiate between Components CAs and User CAs
            cert_der_b64 = element.findtext("tsl:ServiceInformation/tsl:ServiceDigitalIdentity/tsl:DigitalId/tsl:X509Certificate", namespaces=ns)
            cert_der = base64.b64decode(cert_der_b64)
            certificate = x509.load_der_x509_certificate(cert_der, default_backend())
            pem_certificate = certificate.public_bytes(encoding=serialization.Encoding.PEM)
            out.write(pem_certificate)



from typing import List, Tuple
import xml.etree.ElementTree as ET


def element_to_obj(element: ET.Element, single_elements: List[str] = [], collapse_elements: List[Tuple] = []):
    result = {}
    result.update(element.attrib)
    tag = element.tag.split('}')[-1]

    for child in element:
        key = child.tag.split('}')[-1]
        val = element_to_obj(child, single_elements, collapse_elements)
        if key in single_elements:
            result[key] = val
        elif key in result:
            result[key].append(val)
        elif type(val) == str:
            result[key] = val
        else:
            result[key] = [val]

    if len(result.keys()) == 0:
        return element.text

    if element.text is not None:
        result['_text'] = element.text

    for collapse in collapse_elements:
        if collapse[0] == tag:
            return result[key]

    return result



from koap.client import ConnectorClient
from koap.facade.model import (
    CardTypeEnum,
    Card,
    obj_to_card,
    CryptEnum,
    CertRefEnum,
)
from typing import List

# cards which can be used for (external) authentication
AUTH_CARDS = [CardTypeEnum.HBA, CardTypeEnum.SMC_B, CardTypeEnum.HSM_B, CardTypeEnum.SM_B]


class ExternalAuthenticateFacade:
    def __init__(self, client: ConnectorClient):
        self.client = client
        self.event_service = client.create_service_client('EventService', '7.2.0')
        self.card_service = client.create_service_client('CardService', '8.1.2')
        self.certificate_service = client.create_service_client('CertificateService', '6.0.1')
        self.auth_signature_service = client.create_service_client('AuthSignatureService', '7.4.1')

    def get_auth_cards(self) -> List[Card]:
        get_cards_response = self.event_service.GetCards(
            Context=self.client.context()
        )
        raw_cards = filter(lambda c: c.CardType in AUTH_CARDS, get_cards_response.Cards.Card)
        return list(map(obj_to_card, raw_cards))

    def get_card_certificates(self, card_handle: str, cert_types: List[CertRefEnum], crypt: CryptEnum):
        response = self.certificate_service.ReadCardCertificate(
            CardHandle=card_handle,
            Context=self.client.context(),
            CertRefList=list(map(lambda cert_type: cert_type.value, cert_types)),
            Crypt=crypt.value,
        )

        return response.X509DataInfoList.X509DataInfo

    def verify_pin(self, card_type: CardTypeEnum, card_handle: str) -> any:
        if card_type == CardTypeEnum.SMC_B:
            pin_type = "PIN.SMC"
        else:
            pin_type = "PIN.CH"
        return self.card_service.VerifyPin(
            self.client.context(),
            card_handle,
            pin_type
        )

    def external_authenticate(self, card_handle: str, hash: bytes, crypt: CertRefEnum) -> bytes:
        if crypt == "RSA":
            optional_outputs = {
                'SignatureType': 'urn:ietf:rfc:3447',
                'SignatureSchemes': 'RSASSA-PSS'
            }
        else:
            optional_outputs = {
                'SignatureType': 'urn:bsi:tr:03111:ecdsa',
            }

        response = self.auth_signature_service.ExternalAuthenticate(
            CardHandle=card_handle,
            Context=self.client.context(),
            OptionalInputs=optional_outputs,
            BinaryString={
                "Base64Data": {
                    "MimeType": "application/octet-stream",
                    "_value_1": hash
                }
            }
        )

        return response.SignatureObject.Base64Signature._value_1



from .config import ConnectorConfig
from .service_directory import load_service_directory, Service, ServiceVersion
from zeep import Settings, Transport, Client
from zeep.proxy import ServiceProxy
from requests import Session
from requests.auth import HTTPBasicAuth
from enum import Enum
from importlib import resources
from requests_pkcs12 import Pkcs12Adapter
import ssl


def reduce_version(version):
    # Split the version string into parts
    parts = version.split(".")

    # Check if the version string is valid
    if len(parts) != 3:
        raise ValueError("Invalid version string")

    # Return the version without the patch version
    return ".".join(parts[0:2])


def binding_name(sds_service: Service, sds_service_version: ServiceVersion, binding_local_name: str) -> str:
    if binding_local_name is None:
        binding_local_name = f"{sds_service.Name}Binding"
    return f"{{{sds_service_version.TargetNamespace}}}{binding_local_name}"


class ConnectorServiceName(str, Enum):
    SignatureService = 'SignatureService'
    AuthSignatureService = 'AuthSignatureService'
    EncryptionService = 'EncryptionService'
    CardTerminalService = 'CardTerminalService'
    CardService = 'CardService'
    EventService = 'EventService'
    CertificateService = 'CertificateService'
    AMTSService = 'AMTSService'
    PHRManagementService = 'PHRManagementService'
    PHRService = 'PHRService'
    DPEService = 'DPEService'
    NFDService = 'NFDService'
    KVKService = 'KVKService'
    VSDService = 'VSDService'


def wsdl_location(sds_service: Service, sds_service_version: ServiceVersion, module: str) -> str:
    base_dir = resources.files("koap") / "api-telematik" / "conn"
    if module is not None:
        base_dir = base_dir / module
    wsdl_with_version = base_dir / f"{sds_service.Name}_v{sds_service_version.Version.replace('.', '_')}.wsdl"
    if wsdl_with_version.exists():
        return str(wsdl_with_version)
    else:
        wsdl = base_dir / f"{sds_service.Name}.wsdl"
        if not wsdl.exists():
            raise Exception(f"Unable to find WSDL for Service: {sds_service.Name}")
        return str(wsdl)


class ConnectorClient:
    def __init__(self, config: ConnectorConfig, soap_plugins=None):
        self.config = config
        self.soap_plugins = soap_plugins
        session = Session()
        # session.verify = config.trustchain
        if not config.danger_verify_tls:
            session.verify = False
        self.service_directory = load_service_directory(config, session)
        self.soap_settings = Settings()
        # self.soap_settings.strict = False
        self.soap_settings.forbid_entities = False
        if config.auth_method == 'basic':
            session.auth = HTTPBasicAuth(
                self.config.auth_basic_username,
                self.config.auth_basic_password.get_secret_value()
            )
        elif config.auth_method == 'cert':
            adapter = Pkcs12Adapter(
                pkcs12_filename=self.config.auth_cert_p12_filename,
                pkcs12_password=self.config.auth_cert_p12_password.get_secret_value(),
                ssl_protocol=ssl.PROTOCOL_TLS_CLIENT
            )
            session.mount('https://', adapter)
        self.transport = Transport(session=session)

    def create_service_client(
            self,
            service_name: str,
            service_version: str,
            module: str = None,
            binding_local_name: str = None,
            ) -> ServiceProxy:

        sds_service, sds_service_version = self.service_directory.find_service_version(service_name, service_version)

        client = Client(
            wsdl=wsdl_location(sds_service, sds_service_version, module),
            settings=self.soap_settings,
            transport=self.transport,
            plugins=self.soap_plugins
        )

        service = client.create_service(
            binding_name(sds_service, sds_service_version, binding_local_name),
            address=sds_service_version.EndpointTLS.Location
        )

        return service

    def context(self):
        ctx = {
            'MandantId': self.config.mandant_id,
            'ClientSystemId': self.config.client_system_id,
            'WorkplaceId': self.config.workplace_id,
        }
        if self.config.user_id is not None:
            ctx['UserId'] = self.config.user_id
        return ctx



from pydantic import BaseModel
from koap.config import ConnectorConfig
from typing import List, Tuple, Optional
from requests import Session
import xml.etree.ElementTree as ET
from koap.util import element_to_obj


class LocalProductVersion(BaseModel):
    HWVersion: str
    FWVersion: str


class ProductVersion(BaseModel):
    Local: LocalProductVersion


class ProductIdentification(BaseModel):
    ProductVendorID: str
    ProductCode: str
    ProductVersion: ProductVersion


class ProductTypeInformation(BaseModel):
    ProductType: str
    ProductTypeVersion: str


class ProductInformation(BaseModel):
    ProductTypeInformation: ProductTypeInformation
    ProductIdentification: ProductIdentification


class Endpoint(BaseModel):
    Location: str


class ServiceVersion(BaseModel):
    TargetNamespace: str
    Version: str
    EndpointTLS: Endpoint
    Endpoint: Optional[Endpoint] = None


class Service(BaseModel):
    Name: str
    Versions: List[ServiceVersion]


class ServiceInformation(BaseModel):
    Service: List[Service]


class ConnectorServices(BaseModel):
    ProductInformation: ProductInformation
    ServiceInformation: ServiceInformation

    def find_service_version(self, service_name: str, service_version: str) -> Tuple[Service, ServiceVersion]:
        for service in self.ServiceInformation.Service:
            if service.Name == service_name:
                for version in service.Versions:
                    if version.Version == service_version:
                        return (service, version)
                raise Exception(f"Version '{service_version}' for service '{service_name}' is provided by your connector. Available Versions: {','.join(s.Version for s in service.Versions)} ")

        raise Exception(f"Service not found: '{service_name}'")


def load_service_directory(config: ConnectorConfig, session: Session) -> ConnectorServices:
    url = config.construct_url("/connector.sds")

    response = session.get(url)
    root = ET.fromstring(response.text)

    root_obj = element_to_obj(
        root,
        single_elements=[
            "ProductInformation",
            "ProductTypeInformation",
            "ProductIdentification",
            "ProductMiscellaneous",
            "ProductVersion",
            "ServiceInformation",
            "Versions",
            "Local",
            "Endpoint",
            "EndpointTLS",
        ],
        collapse_elements=[
            ("Versions", "Version")
        ]
    )

    # import json
    # print(json.dumps(root_obj, indent=2))

    return ConnectorServices.parse_obj(root_obj)



import asyncio
import socket
import logging
import ssl
import xmltodict
from requests_pkcs12 import create_sslcontext
from koap.config import ConnectorConfig

logger = logging.getLogger(__name__)


class CETP:
    def __init__(self, config: ConnectorConfig, on_event: callable):
        self.config = config
        self.on_event = on_event

    async def _handle_client(self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter):
        magic = await reader.read(4)
        if magic != b'CETP':
            raise Exception('Invalid protocol')
        len_bytes = await reader.read(4)
        message_len = int.from_bytes(len_bytes, byteorder='big')
        event = await reader.read(message_len)
        logger.debug(f'Received event: {event}')
        writer.close()
        self.on_event(xmltodict.parse(event.decode()))

    def _ssl_context(self):
        ssl_ctx = create_sslcontext(
            pkcs12_data=open(self.config.auth_cert_p12_filename, 'rb').read(),
            pkcs12_password_bytes=self.config.auth_cert_p12_password.get_secret_value().encode(),
            ssl_protocol=ssl.PROTOCOL_TLS_SERVER
        )
        ssl_ctx.check_hostname = False
        ssl_ctx.verify_mode = False
        return ssl_ctx
    
    async def start_server(self, host: str = None, port: int = 12201) -> asyncio.Server:
        if host is None:
            hostname = socket.gethostname()
            host = socket.gethostbyname(hostname)

        self._server = await asyncio.start_server(
            self._handle_client, host, port,
            ssl=self._ssl_context()
        )

        coroutine = self._server.serve_forever()
        logger.debug(f'Serving on {host}:{port}')

        return coroutine

    async def stop_server(self):
        self._server.close()
        await self._server.wait_closed()

